

/*  ml16p.c
 *  Copyright 1995 1996, Sebastian Kuzminsky <kuzminsk@cs.colorado.edu>.
 *  See the file "doc/COPYING" for detailed copyright information.
 *
 *     This is the Linux ML16-P driver.
 */


#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/timer.h>

#include <linux/fcntl.h>

#include <asm/segment.h>
#include <asm/io.h>


#include "ml16p-driver.h"

#include "ml16p.h"




#ifdef ML16P_DEBUG
   #define DPRINTK1(a)          printk(ML16P_ID a)
   #define DPRINTK2(a, b)       printk(ML16P_ID a, b)
   #define DPRINTK3(a, b, c)    printk(ML16P_ID a, b, c)
   #define DPRINTK4(a, b, c, d) printk(ML16P_ID a, b, c, d)
#else
   #define DPRINTK1(a)
   #define DPRINTK2(a, b)
   #define DPRINTK3(a, b, c)
   #define DPRINTK4(a, b, c, d)
#endif




/*  Global structures for card info.                                         */
static struct ml16p_data ml16p_card[ML16P_MAX_CARDS];
static int ml16p_num_cards = 0;




static int ml16p_read(
   struct inode *inode,
   struct file *file,
   char *buffer,
   int count
) {
   int i;
   int cardnum;
   unsigned char c;

   unsigned int minor;


   cardnum = ML16P_CARD_NUM(MINOR(inode->i_rdev));
   minor = MINOR(inode->i_rdev) - ML16P_MINOR_START(cardnum);


   if(minor == ML16P_MINOR_DIGITAL) {
      for(i = 0; i < count; i ++) {
	 c = ML16P_DIGITAL_GET(cardnum);
	 put_fs_byte(c, buffer);

	 DPRINTK3("card %c, digital lines, read 0x%02x\n", 'A' + cardnum, c);

	 buffer ++;
      }

      return(count);
   }


   if(minor >= ML16P_MINOR_CT0) {
      if(count != 2) {
         return -ENODATA;
      }

      minor -= ML16P_MINOR_CT0;

      outb((minor << 6) | ML16P_CT_LATCH_CMD, ml16p_card[ml16p_num_cards].baseaddress + ML16P_CT_CONTROL);
      put_fs_byte(ML16P_CT_GET(minor, cardnum), buffer);
      put_fs_byte(ML16P_CT_GET(minor, cardnum), buffer + 1);

      return(count);
   }


   if(minor <= ML16P_MINOR_ADC15) {
      down(&ml16p_card[cardnum].sem);

      if(count == 1) {
/*  Disable automatic conversion start, select ADC, read the one value.      */
	 DPRINTK3("card %c, adc %d, disabling ATO\n", 'A' + cardnum, minor);
	 ml16p_card[cardnum].adcmode[minor] &= 0xff - ML16P_ATO;
         ML16P_ADC_SELECT(minor, cardnum);

	 DPRINTK3("card %c, adc %d, clearing interrupt\n", 'A' + cardnum, minor);
	 ML16P_ADC_RESET_INT_REQ(cardnum);

	 DPRINTK1("entering critical code\n");
	 DPRINTK3("card %c, adc %d, starting single conversion\n", 'A' + cardnum, minor);
         cli();
	 ML16P_ADC_START_CONVERSION(cardnum);

	 sleep_on(&ml16p_card[cardnum].wait_q);
         DPRINTK1("out of critical code\n");
         DPRINTK3("card %c, adc %d woken up\n", 'A' + cardnum, minor);

	 put_fs_byte(ML16P_ADC_GET(cardnum), buffer);
      }
      else {
/*  Enable automatic conversion start, read the whole bunch.                 */
	 DPRINTK3("card %c, adc %d, enabling ATO\n", 'A' + cardnum, minor);
	 ml16p_card[cardnum].adcmode[minor] |= ML16P_ATO;
         ML16P_ADC_SELECT(minor, cardnum);

	 DPRINTK3("card %c, adc %d, clearing interrupt\n", 'A' + cardnum, minor);
	 ML16P_ADC_RESET_INT_REQ(cardnum);

	 DPRINTK1("entering critical code\n");
	 DPRINTK3("card %c, adc %d, starting initial conversion\n", 'A' + cardnum, minor);
         cli();
	 ML16P_ADC_START_CONVERSION(cardnum);

	 for(i = 0; i < count - 1; i ++) {
	    sleep_on(&ml16p_card[cardnum].wait_q);
	    DPRINTK3("card %c, adc %d woken up, copying value to user space\n", 'A' + cardnum, minor);

            cli();
	    put_fs_byte(ML16P_ADC_GET(cardnum), buffer);

	    buffer ++;
	 }
	 sleep_on(&ml16p_card[cardnum].wait_q);
	 DPRINTK3("card %c, adc %d woken up\n", 'A' + cardnum, minor);

	 DPRINTK2("card %c, disabling ATO\n", 'A' + cardnum);
	 ml16p_card[cardnum].adcmode[minor] &= 0xff - ML16P_ATO;
	 ML16P_ADC_SELECT(minor, cardnum);

	 DPRINTK3("card %c, adc %d copying last value to user space\n", 'A' + cardnum, minor);
	 put_fs_byte(ML16P_ADC_GET(cardnum), buffer);
      }

      up(&ml16p_card[cardnum].sem);

      return(count);
   }

   printk(ML16P_ID " bungled minor number %d\n", minor + (cardnum * ML16P_MINOR_RANGE));
   return(0);
}




static int ml16p_write(
   struct inode *inode,
   struct file *file,
   const char *buffer,
   int count
) {
   int cardnum;
   int i;
   unsigned char val;
#ifdef  ML16P_DEBUG
   unsigned short counter_value;
#endif

   unsigned int minor = MINOR(inode->i_rdev);


   cardnum = ML16P_CARD_NUM(minor);
   minor -= ML16P_MINOR_START(cardnum);


   if(minor == ML16P_MINOR_DIGITAL) {
      for(i = 0; i < count; i ++) {
         ML16P_DIGITAL_SET(get_fs_byte(buffer), cardnum);
	 DPRINTK3("card %c, wrote 0x%02x to digital lines\n", 'A' + cardnum, ml16p_card[cardnum].diglines);

	 buffer ++;
      }

      return count;
   }


   if(minor >= ML16P_MINOR_CT0) {
      if(count != 2) {
         return -ENOSPC;
      }

      minor -= ML16P_MINOR_CT0;

      outb(ml16p_card[ml16p_num_cards].ctmode[minor] | ML16P_CT_W_LSB_MSB, ml16p_card[ml16p_num_cards].baseaddress + ML16P_CT_CONTROL);

      val = get_fs_byte(buffer);
#ifdef  ML16P_DEBUG
      counter_value = val;
#endif
      ML16P_CT_SET(val, minor, cardnum);

      buffer ++;
      val = get_fs_byte(buffer);
      ML16P_CT_SET(val, minor, cardnum);
#ifdef  ML16P_DEBUG
      counter_value += (unsigned short)val << 8;
      DPRINTK4("card %c, c/t %d, wrote 0x%04hx\n", 'A' + cardnum, minor, counter_value);
#endif

      return(count);
   }


   if((minor == ML16P_MINOR_DAC0) || (minor == ML16P_MINOR_DAC1)) {
      minor -= ML16P_MINOR_DAC0;
      for(i = 0; i < count; i ++) {
	 ML16P_DAC_SET(get_fs_byte(buffer), minor, cardnum);
	 DPRINTK4("card %c, dac %d, wrote 0x%02x\n", 'A' + cardnum, minor, (unsigned char)get_fs_byte(buffer));

	 buffer ++;
      }

      return(count);
   }


/*  To make the compiler happy:                                              */
   return(0);
}




static int ml16p_ioctl(
   struct inode *inode,
   struct file *file,
   unsigned int command,
   unsigned long argument
) {
   int minor = MINOR(inode->i_rdev);

   int cardnum;


   cardnum = ML16P_CARD_NUM(minor);
   minor -= ML16P_MINOR_START(cardnum);


   switch(command) {
      case ML16P_IO_OFFSET:
         if(minor <= ML16P_MINOR_ADC15) {
	    DPRINTK3("card %c, adc %d ioctl, offset output format\n", 'A' + cardnum, minor);

	    ml16p_card[cardnum].adcmode[minor] &= 0xff - ML16P_FRM;

	    DPRINTK4("card %c, adc %d mode now 0x%02x\n", 'A' + cardnum, minor, ml16p_card[cardnum].adcmode[minor]);
         }
         else {
            printk(ML16P_ID "bungled minor number %d in ioctl OFFSET\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }

         break;


      case ML16P_IO_TWOS_COMP:
         if(minor <= ML16P_MINOR_ADC15) {
	    DPRINTK3("card %c, adc %d ioctl, 2's complement output format\n", 'A' + cardnum, minor);

	    ml16p_card[cardnum].adcmode[minor] |= ML16P_FRM;

	    DPRINTK4("card %c, adc %d mode now 0x%02x\n", 'A' + cardnum, minor, ml16p_card[cardnum].adcmode[minor]);
         }
         else {
            printk(ML16P_ID "bungled minor number %d in ioctl TWOS_COMP\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }

         break;


      case ML16P_IO_UNIPOLAR:
         if(minor <= ML16P_MINOR_ADC15) {
	    DPRINTK3("card %c, adc %d ioctl, unipolar input range\n", 'A' + cardnum, minor);

	    ml16p_card[cardnum].adcmode[minor] &= 0xff - ML16P_BU;

	    DPRINTK4("card %c, adc %d mode now 0x%02x\n", 'A' + cardnum, minor, ml16p_card[cardnum].adcmode[minor]);

         }
         else {
            printk(ML16P_ID "bungled minor number %d in ioctl UNIPOLAR\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }

         break;


      case ML16P_IO_BIPOLAR:
         if(minor <= ML16P_MINOR_ADC15) {
	    DPRINTK3("card %c, adc %d ioctl, bipolar input range\n", 'A' + cardnum, minor);

	    ml16p_card[cardnum].adcmode[minor] |= ML16P_BU;

	    DPRINTK4("card %c, adc %d mode now 0x%02x\n", 'A' + cardnum, minor, ml16p_card[cardnum].adcmode[minor]);
         }
         else {
            printk(ML16P_ID "bungled minor number %d in ioctl BIPOLAR\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }

         break;


      case ML16P_IO_LOW_RANGE:
         if(minor <= ML16P_MINOR_ADC15) {
	    DPRINTK3("card %c, adc %d ioctl, go to low input range\n", 'A' + cardnum, minor);

	    ml16p_card[cardnum].adcmode[minor] &= 0xff - ML16P_RGE;

	    DPRINTK4("card %c, adc %d mode now 0x%02x\n", 'A' + cardnum, minor, ml16p_card[cardnum].adcmode[minor]);
         }
         else {
            printk(ML16P_ID "bungled minor number %d in ioctl LOW_RANGE\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }

         break;


      case ML16P_IO_HIGH_RANGE:
         if(minor <= ML16P_MINOR_ADC15) {
	    DPRINTK3("card %c, adc %d ioctl, go to high input range\n", 'A' + cardnum, minor);

	    ml16p_card[cardnum].adcmode[minor] |= ML16P_RGE;

	    DPRINTK4("card %c, adc %d mode now 0x%02x\n", 'A' + cardnum, minor, ml16p_card[cardnum].adcmode[minor]);
         }
         else {
            printk(ML16P_ID "bungled minor number %d in ioctl HIGH_RANGE\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }

         break;


      case ML16P_IO_SETBIT:
         if(minor == ML16P_MINOR_DIGITAL) {
            int bitnum = (argument & 0x70) >> 4;
            unsigned char val = (argument & 0x01);

            DPRINTK4("card %c, setting bit %d of digital lines to %d\n", 'A' + cardnum, bitnum, val);

	    cli();

	    if(val == 0x01) {
	       ml16p_card[cardnum].diglines |= (0x01 << bitnum);
	    }
	    else {
	       ml16p_card[cardnum].diglines &= 0xff - (0x01 << bitnum);
	    }

	    ML16P_DIGITAL_SET(ml16p_card[cardnum].diglines, cardnum);

	    sti();

	    break;
	 }
	 else {
            printk(ML16P_ID "bungled minor number %d in ioctl SETBIT\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }


      case ML16P_IO_CT_MODE:
         if(minor < ML16P_MINOR_CT0) {
            printk(ML16P_ID "bungled minor number %d in ioctl ML16P_IO_CT_MODE\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }

         DPRINTK4("card %d, c/t %d, switching to mode %ld\n", cardnum, minor - ML16P_MINOR_CT0, argument);

         DPRINTK4("card %d, c/t %d, original mode is 0x%02x\n", cardnum, minor - ML16P_MINOR_CT0, ml16p_card[cardnum].ctmode[minor - ML16P_MINOR_CT0]);
	 ml16p_card[cardnum].ctmode[minor - ML16P_MINOR_CT0] &= (0xff - ML16P_CT_MODE_MASK);
	 DPRINTK4("card %d, c/t %d, cleared mode is 0x%02x\n", cardnum, minor - ML16P_MINOR_CT0, ml16p_card[cardnum].ctmode[minor - ML16P_MINOR_CT0]);

	 switch(argument) {
	    case 0:
	       ml16p_card[cardnum].ctmode[minor - ML16P_MINOR_CT0] |= ML16P_CT_MODE_0;
	       break;

	    case 1:
	       ml16p_card[cardnum].ctmode[minor - ML16P_MINOR_CT0] |= ML16P_CT_MODE_1;
	       break;

	    case 2:
	       ml16p_card[cardnum].ctmode[minor - ML16P_MINOR_CT0] |= ML16P_CT_MODE_2;
	       break;

	    case 3:
	       ml16p_card[cardnum].ctmode[minor - ML16P_MINOR_CT0] |= ML16P_CT_MODE_3;
	       break;

	    case 4:
	       ml16p_card[cardnum].ctmode[minor - ML16P_MINOR_CT0] |= ML16P_CT_MODE_4;
	       break;

	    case 5:
	       ml16p_card[cardnum].ctmode[minor - ML16P_MINOR_CT0] |= ML16P_CT_MODE_5;
	       break;

	    default:
               DPRINTK2("invalid counter/timer mode %ld\n", argument);
               break;
	 }

	 outb(ml16p_card[cardnum].ctmode[minor - ML16P_MINOR_CT0] | ML16P_CT_W_LSB_MSB, (ml16p_card[cardnum].baseaddress + ML16P_CT_CONTROL));
         outb(0xff, ml16p_card[cardnum].baseaddress + ML16P_CT0_WRITE + (minor - ML16P_MINOR_CT0));
         outb(0xff, ml16p_card[cardnum].baseaddress + ML16P_CT0_WRITE + (minor - ML16P_MINOR_CT0));

	 DPRINTK4("card %d, c/t %d, final mode is 0x%02x\n", cardnum, minor - ML16P_MINOR_CT0, ml16p_card[cardnum].ctmode[minor - ML16P_MINOR_CT0]);
         break;


      case  ML16P_IO_INFO:
         if(minor <= ML16P_MINOR_ADC15) {
	    put_fs_byte(ml16p_card[cardnum].adcmode[minor], (unsigned char *)argument);
            break;
         }

         if(minor >= ML16P_MINOR_CT0) {
            DPRINTK4("card %d, c/t %d, mode is 0x%02x\n", cardnum, minor - ML16P_MINOR_CT0, ml16p_card[cardnum].ctmode[minor - ML16P_MINOR_CT0]);
	    put_fs_byte(ml16p_card[cardnum].ctmode[minor - ML16P_MINOR_CT0], (unsigned char *)argument);
            break;
         }

	 printk(ML16P_ID "bungled minor number %d in ioctl INFO\n", MINOR(inode->i_rdev));
	 return -ENOTTY;


#ifdef ML16P_DEBUG
      case ML16P_DEBUG_IO_EN_AUTOCON:
         if(minor <= ML16P_MINOR_ADC15) {
	    printk(ML16P_ID "card %c, adc %d, debug ioctl enable autoconversion\n", 'A' + cardnum, minor);

	    ml16p_card[cardnum].adcmode[minor] |= ML16P_ATO;
	    printk(ML16P_ID "card %c, adc %d mode now 0x%02x\n", 'A' + cardnum, minor, ml16p_card[cardnum].adcmode[minor]);
         }
         else {
            printk(ML16P_ID "bungled minor number %d in debug ioctl EN_AUTOCONVERT\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }

         break;


      case ML16P_DEBUG_IO_DIS_AUTOCON:
         if(minor <= ML16P_MINOR_ADC15) {
	    printk(ML16P_ID "card %c, adc %d, debug ioctl disable autoconversion\n", 'A' + cardnum, minor);

	    ml16p_card[cardnum].adcmode[minor] &= 0xff - ML16P_ATO;
	    printk(ML16P_ID "card %c, adc %d mode now 0x%02x\n", 'A' + cardnum, minor, ml16p_card[cardnum].adcmode[minor]);
         }
         else {
            printk(ML16P_ID "bungled minor number %d in debug ioctl DIS_AUTOCONVERT\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }

         break;


      case ML16P_DEBUG_IO_RIR:
         if(minor <= ML16P_MINOR_ADC15) {
	    printk(ML16P_ID "card %c, adc %d, debug ioctl RIR\n", 'A' + cardnum, minor);
	    outb(0x00, ml16p_card[cardnum].baseaddress + ML16P_RESET_INT_REQ);
         }
         else {
            printk(ML16P_ID "bungled minor number %d in debug ioctl RIR\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }

	 break;


      case ML16P_DEBUG_IO_ADC_START:
         if(minor <= ML16P_MINOR_ADC15) {
	    printk(ML16P_ID "card %c, adc %d, debug ioctl ADC_START\n", 'A' + cardnum, minor);
	    outb(0x00, ml16p_card[cardnum].baseaddress + ML16P_ADC_START);
         }
         else {
            printk(ML16P_ID "bungled minor number %d in debug ioctl ADC_START\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }

	 break;


      case ML16P_DEBUG_IO_ADC_READ:
         if(minor <= ML16P_MINOR_ADC15) {
	    printk(ML16P_ID "card %c, adc %d, debug ioctl ADC_READ\n", 'A' + cardnum, minor);
	    printk(ML16P_ID "card %c, adc %d read 0x%02x\n", 'A' + cardnum, minor, inb(ml16p_card[cardnum].baseaddress + ML16P_ADC_IN));
         }
         else {
            printk(ML16P_ID "bungled minor number %d in debug ioctl ADC_READ\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }

	 break;


      case ML16P_DEBUG_IO_ADC_SELECT:
         if(minor <= ML16P_MINOR_ADC15) {
	    printk(ML16P_ID "card %c, adc %d, debug ioctl ADC_SELECT\n", 'A' + cardnum, minor);
	    outb(ml16p_card[cardnum].adcmode[minor], ml16p_card[cardnum].baseaddress + ML16P_COMMAND);
         }
         else {
            printk(ML16P_ID "bungled minor number %d in debug ioctl ADC_SELECT\n", MINOR(inode->i_rdev));
            return -ENOTTY;
         }

	 break;
#endif /* ML16P_DEBUG */


      default:
         return -EINVAL;
   }

   return 0;
}




static int ml16p_open(struct inode *inode, struct file *file) {
   unsigned int minor = MINOR(inode->i_rdev);

   int cardnum;


   cardnum = ML16P_CARD_NUM(minor);
   minor -= ML16P_MINOR_START(cardnum);
   if ((minor > ML16P_MINOR_LARGEST) || (cardnum >= ml16p_num_cards)) {
      printk(ML16P_ID "trying to open bad minor number %d\n", MINOR(inode->i_rdev));
      return -ENODEV;
   }


   switch(file->f_flags & O_ACCMODE) {
      case O_RDONLY:
         if((ml16p_card[cardnum].read_in_use & (1L << minor)) != 0) {
#ifdef ML16P_DEBUG
	    printk(ML16P_ID "card %c, ", 'A' + cardnum);
	    if(minor == ML16P_MINOR_DIGITAL) {
	       printk("digital lines ");
	    }
	    else if(minor < ML16P_MINOR_DIGITAL) {
	       printk("analog line %d ", minor);
	    }
	    else {
	       printk("counter/timer %d ", minor - ML16P_MINOR_CT0);
	    }
	    printk("already opened for reading\n");
#endif
	    return -EBUSY;
	 }

	 ml16p_card[cardnum].read_in_use |= (1L << minor);

#ifdef ML16P_DEBUG
	 printk(ML16P_ID "card %c, ", 'A' + cardnum);
	 if(minor == ML16P_MINOR_DIGITAL) {
	    printk("digital lines ");
	 }
	 else if(minor < ML16P_MINOR_DIGITAL) {
	    printk("analog line %d ", minor);
	 }
	 else {
	    printk("counter/timer %d ", minor - ML16P_MINOR_CT0);
	 }
	 printk("opened for reading\n");
#endif

         break;


      case O_WRONLY:
         if ((minor != ML16P_MINOR_DAC0) && (minor != ML16P_MINOR_DAC1) &&  \
            (minor != ML16P_MINOR_DIGITAL) && (minor != ML16P_MINOR_CT0) &&  \
            (minor != ML16P_MINOR_CT1) && (minor != ML16P_MINOR_CT2)) {

	    DPRINTK3("card %c, analog line %d cannot be written to\n", 'A' + cardnum, minor);

            return -EPERM;
         }

	 if((ml16p_card[cardnum].write_in_use & (1L << minor)) != 0) {
#ifdef ML16P_DEBUG
	    printk(ML16P_ID "card %c, ", 'A' + cardnum);
	    if(minor == ML16P_MINOR_DIGITAL) {
	       printk("digital lines ");
	    }
	    else if(minor < ML16P_MINOR_DIGITAL) {
	       printk("analog line %d ", minor);
	    }
	    else {
	       printk("counter/timer %d ", minor - ML16P_MINOR_CT0);
	    }
	    printk("already opened for writing\n");
#endif
	    return -EBUSY;
	 }

	 ml16p_card[cardnum].write_in_use |= (1L << minor);

#ifdef ML16P_DEBUG
	 printk(ML16P_ID "card %c, ", 'A' + cardnum);
	 if(minor == ML16P_MINOR_DIGITAL) {
	    printk("digital lines ");
	 }
	 else if(minor < ML16P_MINOR_DIGITAL) {
	    printk("analog line %d ", minor);
	 }
	 else {
	    printk("counter/timer %d ", minor - ML16P_MINOR_CT0);
	 }
	 printk("opened for writing\n");
#endif

	 break;


      case O_RDWR:
         if ((minor != ML16P_MINOR_DIGITAL) && (minor != ML16P_MINOR_CT0) &&  \
            (minor != ML16P_MINOR_CT1) && (minor != ML16P_MINOR_CT2)) {

	    DPRINTK3("card %c, analog line %d cannot be written to\n", 'A' + cardnum, minor);

            return -EPERM;
         }

	 if (((ml16p_card[cardnum].write_in_use & (1L << minor)) != 0) || \
	    ((ml16p_card[cardnum].read_in_use & (1L << minor)) != 0)) {
#ifdef ML16P_DEBUG
	    printk(ML16P_ID "card %c, ", 'A' + cardnum);
	    if(minor == ML16P_MINOR_DIGITAL) {
	       printk("digital lines ");
	    }
	    else if(minor < ML16P_MINOR_DIGITAL) {
	       printk("analog line %d ", minor);
	    }
	    else {
	       printk("counter/timer %d ", minor - ML16P_MINOR_CT0);
	    }
	    printk("already opened for reading and/or writing\n");
#endif
	    return -EBUSY;
	 }

	 ml16p_card[cardnum].read_in_use |= (1L << minor);
	 ml16p_card[cardnum].write_in_use |= (1L << minor);

#ifdef ML16P_DEBUG
	 printk(ML16P_ID "card %c, ", 'A' + cardnum);
	 if(minor == ML16P_MINOR_DIGITAL) {
	    printk("digital lines ");
	 }
	 else if(minor < ML16P_MINOR_DIGITAL) {
	    printk("analog line %d ", minor);
	 }
	 else {
	    printk("counter/timer %d ", minor - ML16P_MINOR_CT0);
	 }
	 printk("opened for reading and writing\n");
#endif

	 break;
   }
	 
/*  only adc0-7 are available for input when card is configured for
 *  Differential operation
 */

   MOD_INC_USE_COUNT;

   return 0;
}




static void ml16p_close(struct inode *inode, struct file *file) {
   unsigned int minor = MINOR(inode->i_rdev);
   int cardnum;


   cardnum = ML16P_CARD_NUM(minor);
   minor -= ML16P_MINOR_START(cardnum);


   switch(file->f_flags & O_ACCMODE) {
      case O_RDONLY:
	 ml16p_card[cardnum].read_in_use &= ~(1L << minor);
         break;


      case O_WRONLY:
	 ml16p_card[cardnum].write_in_use &= ~(1L << minor);
	 break;


      case O_RDWR:
	 ml16p_card[cardnum].read_in_use &= ~(1L << minor);
	 ml16p_card[cardnum].write_in_use &= ~(1L << minor);
	 break;
   }

#ifdef ML16P_DEBUG
   printk(ML16P_ID "card %c, ", 'A' + cardnum);
   if(minor == ML16P_MINOR_DIGITAL) {
      printk("digital lines ");
   }
   else if(minor < ML16P_MINOR_DIGITAL) {
      printk("analog line %d ", minor);
   }
   else {
      printk("counter/timer %d ", minor - ML16P_MINOR_CT0);
   }
   printk("closed\n");
#endif

   MOD_DEC_USE_COUNT;
}




static struct file_operations ml16p_fops =
{
   NULL,                   /*  ml16p_lseek               */
   ml16p_read,             /*  ml16p_read                */
   ml16p_write,            /*  ml16p_write               */
   NULL,                   /*  ml16p_readaddr            */
   NULL,                   /*  ml16p_select              */
   ml16p_ioctl,            /*  ml16p_ioctl               */
   NULL,                   /*  ml16p_mmap                */
   ml16p_open,             /*  ml16p_open                */
   ml16p_close,            /*  ml16p_close               */
   NULL,                   /*  ml16p_fsync               */
   NULL,                   /*  ml16p_fasync              */
   NULL,                   /*  ml16p_check_media_change  */
   NULL                    /*  ml16p_revalidate          */
};




void ml16p_probe_wakeup(unsigned long wait_q) {
   DPRINTK1("card probe waker-upper here\n");

   wake_up((struct wait_queue **)wait_q);
}




void ml16p_irq_handler(int irq, void *dev_id, struct pt_regs *regs) {
   unsigned char cardnum;


   cardnum = ml16p_num_cards - 1;

   while(ml16p_card[cardnum].irq != irq) {
      if(cardnum == 0) {
         return;
      }

      cardnum --;
   }

   DPRINTK3("card %c, irq %d, irq handler running\n", 'A' + cardnum, irq);

   wake_up(&ml16p_card[cardnum].wait_q);
}




void ml16p_probe_isr(int irq, void *dev_id, struct pt_regs *regs) {
   DPRINTK2("prober ISR running, recieved irq %d\n", irq);

   ml16p_card[ml16p_num_cards].irq = irq;
}




int init_module(void) {
   unsigned char irq;
   unsigned int irqs = 0;

   unsigned short probe_address;

   struct wait_queue *probe_wait_q;
   struct timer_list probe_timer;

   int i;


   printk(ML16P_ID ML16P_VERSION ", loading...\n");

/*  Register the device with the kernel.                                     */
   DPRINTK2("registering driver, major number %d\n", ML16P_MAJOR);
   if(register_chrdev(ML16P_MAJOR, "ml16p", &ml16p_fops)) {
      printk(ML16P_ID "unable to get major number %d\n", ML16P_MAJOR);
      return -EBUSY;
   }

/*  Grab all free irq's.  Normally one should use probe_irq_on() and         */
/*  probe_irq_off(), but that didnt seem appropriate here.                   */
   DPRINTK1("aquiring irq probe mask\n");
   for(irq = ML16P_PROBE_IRQ_START; irq <= ML16P_PROBE_IRQ_END; irq ++) {
      if(request_irq(irq, ml16p_probe_isr, 0, "ml16p probe", NULL)) {
         DPRINTK2("irq %d not available for probe\n", irq);
      }
      else {
         DPRINTK2("irq %d aquired for probe\n", irq);
         irqs |= (unsigned int)(1 << irq);
      }
   }

/*  Now test to see if there's an ML16-P card at each available I/O address  */
/*  To find out, start an ADC conversion and listen for the                  */
/*  End-of-Conversion interrupt.                                             */
   for(probe_address = ML16P_PROBE_IO_START; probe_address <= ML16P_PROBE_IO_END; probe_address += 0x010) {
      DPRINTK2("checking 0x10 I/O ports starting at 0x%03x\n", probe_address);
      if(check_region(probe_address, 16)) {
         continue;
      }
      DPRINTK1("I/O ports are available\n");

      DPRINTK1("probing...\n");
      ml16p_card[ml16p_num_cards].irq = 0;

      init_timer(&probe_timer);
      probe_timer.expires = ML16P_PROBE_TIME;
      probe_timer.data = (unsigned long )&probe_wait_q;
      probe_timer.function = ml16p_probe_wakeup;
      probe_wait_q = NULL;

/*  Select an ADC, any ADC.                                                  */
      outb(ML16P_RGE, probe_address + ML16P_COMMAND);

/*  Clear the End-of-Conversion interrupt.                                   */
      outb(0x00, probe_address + ML16P_RESET_INT_REQ);

/*  Start timer.                                                             */
      cli();
      add_timer(&probe_timer);

/*  Start conversion.                                                        */
      outb(0x00, probe_address + ML16P_ADC_START);

/*  Will be woken by ml16p_probe_wakeup() in ML16P_PROBE_TIME jiffies.       */
      sleep_on(&probe_wait_q);

/*  ML16P_PROBE_TIME jiffies has passed, if the card is installed at this    */
/*  base address and a free IRQ, then ml16p_probe_isr has set the irq field  */
/*  for this ml16p_card.                                                     */

      del_timer(&probe_timer);

      if(ml16p_card[ml16p_num_cards].irq == 0) {
	 DPRINTK2("no card found at address 0x%03x\n", probe_address);

	 continue;
      }

      DPRINTK4("card %c detected at base address 0x%03x, irq %d\n", 'A' + ml16p_num_cards, probe_address, ml16p_card[ml16p_num_cards].irq);

      ml16p_card[ml16p_num_cards].baseaddress = probe_address;
      ml16p_card[ml16p_num_cards].analog_type = ML16P_SINGLE_ENDED;
      ml16p_card[ml16p_num_cards].sem = MUTEX;
      ml16p_card[ml16p_num_cards].wait_q = NULL;


/*  Re-register the irq to the real IRQ handler.                             */
      DPRINTK2("freeing irq %d from probe ISR\n", ml16p_card[ml16p_num_cards].irq);
      free_irq(ml16p_card[ml16p_num_cards].irq, NULL);
      irqs ^= (unsigned int)(1 << ml16p_card[ml16p_num_cards].irq);

      DPRINTK3("requesting irq %d for real ISR for card %c\n", ml16p_card[ml16p_num_cards].irq, 'A' + ml16p_num_cards);
      if(request_irq(ml16p_card[ml16p_num_cards].irq, ml16p_irq_handler, SA_INTERRUPT, "ml16p", NULL)) {
	 DPRINTK3("unable to get irq %d for card %c\n", ml16p_card[ml16p_num_cards].irq, 'A' + ml16p_num_cards);

/*  we should clean up a little better here... */

	 continue;
      }

/*  Claim the I/O ports.                                                     */
      DPRINTK3("requesting I/O port 0x%03x for card %c\n", ml16p_card[ml16p_num_cards].baseaddress, 'A' + ml16p_num_cards);
      request_region(ml16p_card[ml16p_num_cards].baseaddress, 16, "ml16p");

/*  Initialize the digital output lines.                                     */
      ML16P_DIGITAL_SET(0x00, ml16p_num_cards);

/*  Initialize the ADCs.                                                     */
      for(i = 0x00; i <= 0x0f; i ++) {
	 ml16p_card[ml16p_num_cards].adcmode[i] = ML16P_RGE | i;
	 outb(ml16p_card[ml16p_num_cards].adcmode[i], (ml16p_card[ml16p_num_cards].baseaddress + ML16P_COMMAND));

	 DPRINTK4("card %c, adc %d, mode=0x%02x\n", 'A' + ml16p_num_cards, i, ml16p_card[ml16p_num_cards].adcmode[i]);
      }

/*  Initialize the counter/timers.                                           */
      for(i = 0x00; i <= 0x02; i ++) {
	 ml16p_card[ml16p_num_cards].ctmode[i] = (i << 6) |  ML16P_CT_MODE_2 | ML16P_CT_BIN;
	 outb(ml16p_card[ml16p_num_cards].ctmode[i] | ML16P_CT_W_LSB_MSB, (ml16p_card[ml16p_num_cards].baseaddress + ML16P_CT_CONTROL));
         outb(0xff, ml16p_card[ml16p_num_cards].baseaddress + ML16P_CT0_WRITE + i);
         outb(0xff, ml16p_card[ml16p_num_cards].baseaddress + ML16P_CT0_WRITE + i);

	 DPRINTK4("card %c, c/t %d, mode=0x%02x\n", 'A' + ml16p_num_cards, i, ml16p_card[ml16p_num_cards].ctmode[i]);
      }

      printk(ML16P_ID "card %c initialized (addr=0x%03x, irq=%d)\n", 'A' + ml16p_num_cards, ml16p_card[ml16p_num_cards].baseaddress, ml16p_card[ml16p_num_cards].irq);

      ml16p_num_cards ++;
      if(ml16p_num_cards == ML16P_MAX_CARDS) {
	 probe_address = ML16P_PROBE_IO_END + 0x010;
      }
   }


/*  Free all IRQ's not used by ML16-P cards.                                 */
   for(irq = ML16P_PROBE_IRQ_START; irq <= ML16P_PROBE_IRQ_END; irq ++) {
      if(irqs & (unsigned int)(1 << irq)) {
         DPRINTK2("freeing irq %d\n", irq);
         free_irq(irq, NULL);
      }
   }

   if(ml16p_num_cards == 0) {
      DPRINTK2("unregistering device %d\n", ML16P_MAJOR);
      if(unregister_chrdev(ML16P_MAJOR, "ml16p")) {
	 printk(ML16P_ID "device unregistering failed\n");
      }

      printk(ML16P_ID "no cards found, driver not loaded\n");

      return -EBUSY;
   }

   return 0;
}




void cleanup_module(void) {
   unsigned char cardnum;


   for(cardnum = 0; cardnum < ml16p_num_cards; cardnum ++) {
      DPRINTK3("card %c, releasing io port 0x%03x\n", 'A' + cardnum, ml16p_card[cardnum].baseaddress);
      release_region(ml16p_card[cardnum].baseaddress, 16);

      DPRINTK3("card %c, releasing irq %d\n", 'A' + cardnum, ml16p_card[cardnum].irq);
      free_irq(ml16p_card[cardnum].irq, NULL);

      printk(ML16P_ID "card %c released\n", 'A' + cardnum);
   }

   DPRINTK1("unregistering device\n");
   if(unregister_chrdev(ML16P_MAJOR, "ml16p")) {
      printk(ML16P_ID "device unregistering failed\n");
   } else {
      printk(ML16P_ID "driver unloaded\n");
   }
}


