

/*
 *  range.c
 *  Copyright 1995 1996, Sebastian Kuzminsky <kuzminsk@cs.colorado.edu>.
 *  See the file "doc/COPYING" for detailed copyright information.
 *
 *
 *     This program is a command-line utility intended for distribution and use
 *  with the Linux ML16-P driver.  It can retrieve the current configuration of
 *  an ADC and report it to the user.  It can also change the current
 *  configuration of an ADC.
 *
 */


#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>


#include "../driver/ml16p.h"

#include "support.h"




void printhelp(void);




int main(int argc, char **argv) {
   unsigned char val;

   int devicefile = -1;

   int i;


/*  Parse command line, looking for options.                                 */
   for(i = 1; i < argc; i ++) {
      if((strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "--help") == 0)) {
	 printhelp();

	 exit(0);
      }

      if((strcmp(argv[i], "-v") == 0) || (strcmp(argv[i], "--version") == 0)) {
	 printf("range - " VERSION "\n");

	 exit(0);
      }
   }

/*  Incorrect number of command-line arguments?                              */
   if(argc == 1) { 
      printhelp();
      exit(1);
   }


/*  Since the ADC is a read-only device (the DAC is the write part of the    */
/*  special file), we need only open the specified device as read-only.      */
   if((devicefile = open(argv[1], O_RDONLY)) == -1) {
      fprintf(stderr, "cannot open %s for read\n", argv[1]);
      perror(argv[1]);
      exit(1);
   }

/*  First set the range as per the command line.                             */
   for(val = 2; val < argc; val ++) { 
      switch(argv[val][0]) {   /*  We only care about the first character.   */
         case 'u':
/*  'u' for unipolar...                                                      */
	    if(ioctl(devicefile, ML16P_IO_UNIPOLAR, 0L) == -1) {
	       fprintf(stderr, "cannot switch to unipolar\n");
	       perror("ML16P_IO_UNIPOLAR ioctl");
	       exit(1);
	    }
	    continue;

         case 'b':
/*  'b' for bipolar...                                                       */
	    if(ioctl(devicefile, ML16P_IO_BIPOLAR, 0L) == -1) {
	       fprintf(stderr, "cannot switch to bipolar\n");
	       perror("ML16P_IO_BIPOLAR ioctl");
	       exit(1);
	    }
	    continue;

         case 'h':
/*  'h' for high range.  Not to be confused with "-h" for help.              */
	    if(ioctl(devicefile, ML16P_IO_HIGH_RANGE, 0L) == -1) {
	       fprintf(stderr, "cannot switch to high range\n");
	       perror("ML16P_IO_HIGH_RANGE ioctl");
	       exit(1);
	    }
	    continue;

         case 'l':
/*  'l' for low range...                                                     */
	    if(ioctl(devicefile, ML16P_IO_LOW_RANGE, 0L) == -1) {
	       fprintf(stderr, "cannot switch to low range\n");
	       perror("ML16P_IO_LOW_RANGE ioctl");
	       exit(1);
	    }
	    continue;

	 default:
/*  Anything else we reject, then complain and commit suicide.               */
	    printf("unknown argument: %s\n", argv[val]);
	    exit(1);
      }
   }

/*  Next we get the info back from the ADC...                                */
   if(ioctl(devicefile, ML16P_IO_INFO, &val) == -1) {
      fprintf(stderr, "cannot execute INFO ioctl\n");
      perror("INFO ioctl");
      exit(1);
   }

/*  Parse out the bits we care about (see the header file).                  */
/*  The ML16P_* constants used below to report the input range are also      */
/*  defined in the header file.                                              */
   if(val & ML16P_BU) {
      printf("bipolar, ");
      if(val & ML16P_RGE) {
	 printf("high range\n");
	 printf("%.3f VDC to %.3f VDC\n", ML16P_ADC_BI_HI_MIN, ML16P_ADC_BI_HI_MAX);
      }
      else {
	 printf("low range\n");
	 printf("%.3f VDC to %.3f VDC\n", ML16P_ADC_BI_LO_MIN, ML16P_ADC_BI_LO_MAX);
      }
   }
   else {
      printf("unipolar, ");
      if(val & ML16P_RGE) {
	 printf("high range\n");
	 printf("%.3f VDC to %.3f VDC\n", ML16P_ADC_UNI_HI_MIN, ML16P_ADC_UNI_HI_MAX);
      }
      else {
	 printf("low range\n");
	 printf("%.3f VDC to %.3f VDC\n", ML16P_ADC_UNI_LO_MIN, ML16P_ADC_UNI_LO_MAX);
      }
   }

   close(devicefile);

   exit(0);
}




void printhelp(void) {
   printf("Usage:  range [OPTIONS]\n");
   printf("        range DEVICE [SETTING...]\n");
   printf("Gets or sets the input range of an ADC on an ML16-P I/O card.\n");
   printf("\n");
   printf("   -h,  --help     print this help and exit\n");
   printf("   -v,  --version  print version information and exit\n");
   printf("\n");
   printf("DEVICE is the ADC device to access (for example, /dev/ml16pa-adc0)\n");
   printf("The valid SETTINGs are:  { unipolar bipolar high-range low-range }\n");
   printf("Abbreviations are accepted.\n");
   printf("\n");
   printf("When SETTINGs are given, then the ADC is configured as specified and the new\n");
   printf("configuration and input limits are printed out.  Any number of SETTINGs may be\n");
   printf("given.  Multiple SETTINGs are processed in the order they are found on the\n");
   printf("command line, so in case of conflicting SETTINGs the one last entered takes\n");
   printf("effect.\n");
   printf("\n");
   printf("If no SETTINGs are given, the current configuration and input limits are\n");
   printf("printed out.\n");
   printf("\n");
}


