

/*
 *  volts.c
 *  Copyright 1995 1996, Sebastian Kuzminsky <kuzminsk@cs.colorado.edu>.
 *  See the file "doc/COPYING" for detailed copyright information.
 *
 *
 *     This program is a command-line utility intended for distribution and use
 *  with the Linux ML16-P driver.  It can read data from one of the analog to
 *  digital converters and report the voltage to the user in a human readable
 *  form.  It can also accept a voltage as input and cause that voltage to be
 *  written to one of the digital to analog converters.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>
#include <sys/ioctl.h>


#include "../driver/ml16p.h"

#include "support.h"




void printhelp(void);




int main(int argc, char **argv) {
   unsigned char val;

   float voltage;
   float voltage_min;
   float voltage_max;

   int devicefile = -1;

   int i;


/*  Parse command line, looking for options.                                 */
   for(i = 1; i < argc; i ++) {
      if((strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "--help") == 0)) {
	 printhelp();

	 exit(0);
      }

      if((strcmp(argv[i], "-v") == 0) || (strcmp(argv[i], "--version") == 0)) {
	 printf("volts - " VERSION "\n");

	 exit(0);
      }
   }

/*  Incorrect number of command-line arguments?                              */
   if((argc != 2) && (argc != 3)) {
      printhelp();
      exit(1);
   }


/*  No arguments -- read volts from ADC.                                     */
   if(argc == 2) {
/*  Open the ADC.  Since the ADC is a read-only device (the DAC is the write */
/*  part of the special file) we open it read-only.                          */
      if((devicefile = open(argv[1], O_RDONLY)) == -1) {
	 fprintf(stderr, "cannot open %s for read\n", argv[1]);
	 perror(argv[1]);
	 exit(1);
      }

/*  Get the current input range of the ADC so we know how to interpret the   */
/*  reading.                                                                 */
      if(ioctl(devicefile, ML16P_IO_INFO, &val) == -1) {
         fprintf(stderr, "cannot execute info ioctl\n");
	 exit(1);
      }

/*  Set the variables "voltage_min" and "voltage_max" according to the       */
/*  information returned.  Note the use of predefined bits and voltage       */
/*  values from the header file "sys/ml16p.h".                               */
      if(val & ML16P_BU) {
         if(val & ML16P_RGE) {
            voltage_max = ML16P_ADC_BI_HI_MAX;
            voltage_min = ML16P_ADC_BI_HI_MIN;
         }
         else {
            voltage_max = ML16P_ADC_BI_LO_MAX;
            voltage_min = ML16P_ADC_BI_LO_MIN;
         }
      }
      else {
         if(val & ML16P_RGE) {
            voltage_max = ML16P_ADC_UNI_HI_MAX;
            voltage_min = ML16P_ADC_UNI_HI_MIN;
         }
         else {
            voltage_max = ML16P_ADC_UNI_LO_MAX;
            voltage_min = ML16P_ADC_UNI_LO_MIN;
         }
      }

/*  Now get the actual reading from the device.                              */
      if(read(devicefile, &val, 1) != 1) {
         fprintf(stderr, "read error\n");
         exit(1);
      }

/*  Convert binary reading to a float and print it.                          */
      voltage = (((float)val / 255.0) * (voltage_max - voltage_min)) + voltage_min;
      printf("%.3f\n", voltage);

/*  Technically not needed (exit() does it for you) but it makes the code    */
/*  easier to read.                                                          */
      close(devicefile);

      exit(0);
   }

/*  If we get here then there was an argument and we should set the voltage  */
/*  of a DAC.                                                                */
   voltage_max = ML16P_DAC_MAX;
   voltage_min = ML16P_DAC_MIN;

/*  Desired voltage is the first command-line argument.                      */
   voltage = (float)atof(argv[2]);
   if((voltage < voltage_min) || (voltage > voltage_max)) {
      fprintf(stderr, "voltage %f out of range\n", voltage);
      fprintf(stderr, "acceptable range is %.3f to %.3f\n", voltage_min, voltage_max);
      exit(1);
   }

/*  Convert to the correct binary representation.                            */
   val = (unsigned char)(255.0 * ((voltage - voltage_min) / (voltage_max - voltage_min)));

/*  Note write-only open().                                                  */
   if((devicefile = open(argv[1], O_WRONLY)) == -1) {
      fprintf(stderr, "cannot open %s for write\n", argv[1]);
      perror(argv[1]);
      exit(1);
   }

/*  Write to the device.                                                     */
   if(write(devicefile, &val, 1) != 1) {
      fprintf(stderr, "write error\n");
      perror(argv[1]);
      exit(1);
   }

/*  Technically not needed (exit() does it for you) but it makes the code    */
/*  easier to read.                                                          */
   close(devicefile);

   exit(0);
}




void printhelp(void) {
   printf("Usage:  volts [OPTIONS]\n");
   printf("        volts DEVICE [value]\n");
   printf("Reads voltagets from ADCs and sets voltages on DACs.\n");
   printf("\n");
   printf("   -h,  --help     print this help and exit\n");
   printf("   -v,  --version  print version information and exit\n");
   printf("\n");
   printf("DEVICE is the analog device to access (for example, /dev/ml16pa-adc0)\n");
   printf("value is the voltage to output from that device (only valid with DACs)\n");
   printf("\n");
   printf("If no voltage is given on the command line, then ADC access is assumed, and a\n");
   printf("voltage is read from the given device.  If a voltage is supplied and the given\n");
   printf("device is a DAC, then the supplied voltage is output from the device.\n");
   printf("\n");
}


